@@ -6,10 +6,6 @@
 
 */
 
-#ifdef WIN32
-#include <windows.h>
-#endif
-
 #include <stdlib.h>
 #include <string.h>
 
@@ -498,7 +494,7 @@ void transformSurfaceRGBA(SDL_Surface * src, SDL_Surface * dst, int cx, int cy,
     ay = (cy << 16) - (isin * cx);
     sw = src->w - 1;
     sh = src->h - 1;
-    pc = dst->pixels;
+    pc = (tColorRGBA*)dst->pixels;
     gap = dst->pitch - dst->w * 4;
 
     /*
@@ -654,7 +650,7 @@ void transformSurfaceY(SDL_Surface * src, SDL_Surface * dst, int cx, int cy, int
     ay = (cy << 16) - (isin * cx);
     sw = src->w - 1;
     sh = src->h - 1;
-    pc = dst->pixels;
+    pc = (tColorY*)dst->pixels;
     gap = dst->pitch - dst->w;
     /*
      * Clear surface to colorkey 
