@@ -33,76 +33,46 @@
 #include "input.h"
 #include "sound.h"
 #include "util.h"
+#include "network.h"
 
 
 GameState *gamestate_new() {
-    int x, y, z;
     GameState *s;
 
-    GameState template = {
+    GameState gs_template = {
         NULL,
         -1,
-        { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.6, false, -1, false },
+        { 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, false, -1, false },
         {
-            { NULL, -1, GAME_X_MIN-RACKET_X_MID*2, GAME_Y_MID, 0.0, POWER_UP_FACTOR, POWER_DOWN_FACTOR, true, 0, DESIRE_NORMAL, PLAYER_TYPE_AI, GAME_Y_MID, false, 0, {0}, 0, 0, PLAYER_ACCEL_DEFAULT, true },
-            { NULL, -1, GAME_X_MAX+RACKET_X_MID*2, GAME_Y_MID, 0.0, POWER_UP_FACTOR, POWER_DOWN_FACTOR, true, 0, DESIRE_NORMAL, PLAYER_TYPE_AI, GAME_Y_MID, false, 0, {0}, 0, 0, PLAYER_ACCEL_DEFAULT, true },
+            { NULL, -1, GAME_X_MIN-RACKET_X_MID*2, GAME_Y_MID, 0.0, true, 0, DESIRE_NORMAL, PLAYER_TYPE_AI, 0, {0}, PLAYER_ACCEL_DEFAULT },
+            { NULL, -1, GAME_X_MAX+RACKET_X_MID*2, GAME_Y_MID, 0.0, true, 0, DESIRE_NORMAL, PLAYER_TYPE_AI, 0, {0}, PLAYER_ACCEL_DEFAULT },
         },
-        0,
-        0,
         1,
-        "welcome to tennix " VERSION,
-        { 0 },
-        { 0 },
         REFEREE_NORMAL,
         0,
         WINNER_NONE,
-        false,
-        GR_COUNT,
-        { 0 },
-        0,
-        false,
-        { { { 0 } } },
-        0.0,
-        0, /* sound events */
-        0.0,
-        0.0,
-        0,
-        0,
-        0,
-        false,
-        0,
-        0,
-        0,
-        false,
-        REFEREE_COUNT,
+        SOUND_EVENT_NONE, /* sound events */
         SCORE_UNDECIDED,
         0,
+        EVENTCOUNTER_GAMESTATE_START,
+        EVENTCOUNTER_GAMESTATE_START,
+        STATUSMSG_WELCOME,
     };
 
     s = (GameState*)malloc(sizeof(GameState));
     if (s == NULL) abort();
 
-    memcpy(s, &template, sizeof(GameState));
+    memcpy(s, &gs_template, sizeof(GameState));
 
     game_setup_serve(s);
 
-    /* smoothen n-gram */
-    for( x = 0; x<NGRAM_STEPS; x++) {
-        for( y = 0; y<NGRAM_STEPS; y++) {
-            for( z = 0; z<NGRAM_STEPS; z++) {
-                s->ngram[x][y][z] = 1;
-            }
-        }
-    }
-
     return s;
 }
 
 
 int gamestate_save(GameState* s, const char* filename)
 {
-    Location *location;
-    const char* status;
+    const Location *location;
     InputDevice* input_devices[MAXPLAYERS];
     int i, result = 0;
     FILE* fp = NULL;
@@ -129,8 +99,6 @@ int gamestate_save(GameState* s, const char* filename)
     /* Save data for later recovery + clear */
     location = s->location;
     s->location = NULL;
-    status = s->status;
-    s->status = NULL;
     for (i=1; i<=MAXPLAYERS; i++) {
         input_devices[i-1] = PLAYER(s, i).input;
         PLAYER(s, i).input = NULL;
@@ -142,7 +110,6 @@ int gamestate_save(GameState* s, const char* filename)
 
     /* Restore process-specific data */
     s->location = location;
-    s->status = status;
     for (i=1; i<=MAXPLAYERS; i++) {
         PLAYER(s, i).input = input_devices[i-1];
     }
@@ -175,7 +142,6 @@ GameState* gamestate_load(const char* filename)
         if (s != NULL) {
             if (fread(s, sizeof(GameState), 1, fp) == 1) {
                 /* Restore/create process-specific data */
-                s->status = format_status(s);
                 /* FIXME: s->location, players' "input" */
             } else {
                 free(s);
@@ -193,7 +159,7 @@ GameState* gamestate_load(const char* filename)
 }
 
 
-void gameloop(GameState *s) {
+void gameloop(GameState *s, TennixNet* c) {
     Uint32 ot = SDL_GetTicks();
     Uint32 nt;
     Uint32 dt = GAME_TICKS;
@@ -201,23 +167,31 @@ void gameloop(GameState *s) {
     Uint32 accumulator = 0;
     bool quit = false;
     int p;
+    int i=0;
+    RenderState r = {
+        SOUND_EVENT_NONE,
+        REFEREE_COUNT,
+        EVENTCOUNTER_RENDERSTATE_START,
+        EVENTCOUNTER_RENDERSTATE_START,
+        STATUSMSG_NONE,
+        NULL,
+        NULL,
+        NULL,
+    };
+
+    /* Catch-up with existing sound events */
+    r.sound_events = s->sound_events;
 
 #ifdef ENABLE_FPS_LIMIT
     Uint32 ft, frames; /* frame timer and frames */
 #endif
 
-    strcpy(s->game_score_str, format_game(s));
-    strcpy(s->sets_score_str, format_sets(s));
-    s->text_changed = true;
-
-    if (s->rain > 0) {
+    if (s->location->rain > 0) {
         play_sample_background(SOUND_RAIN);
     }
     if (s->location->max_visitors > 100) {
         play_sample_loop(SOUND_AUDIENCE);
     }
-    /* Reset the court type, so it is redrawn on first display */
-    s->displayed_court_type = GR_COUNT;
 
     for (p=1; p<=MAXPLAYERS; p++) {
         input_device_join_game(PLAYER(s, p).input, s, p);
@@ -238,18 +212,17 @@ void gameloop(GameState *s) {
         ot = nt;
 
         while( accumulator >= dt) {
-            quit = step(s);
-            s->time += dt;
-            s->windtime += s->wind*dt;
-            accumulator -= dt;
+            step(s);
 
-            if( s->was_stopped) {
-                ot = SDL_GetTicks();
-                s->was_stopped = false;
+            network_receive(c);
+            network_get_gamestate(c, s);
+
+            if ((i++) % 10 == 0) {
+                network_send_state(c, s);
             }
-        }
-        if (s->timelimit != 0 && s->time >= s->timelimit) {
-            quit = 1;
+
+            quit = handle_input(s, c);
+            accumulator -= dt;
         }
 
 #ifdef ENABLE_FPS_LIMIT
@@ -259,7 +232,7 @@ void gameloop(GameState *s) {
         frames++;
 #endif
 
-        render(s);
+        render(s, &r);
     }
 
     for (p=1; p<=MAXPLAYERS; p++) {
@@ -267,19 +240,33 @@ void gameloop(GameState *s) {
     }
 
     clear_screen();
-    rectangle(0, 0, WIDTH, HEIGHT, 80, 80, 80);
+    //rectangle(0, 0, WIDTH, HEIGHT, 80, 80, 80);
     store_screen();
 
     stop_sample(SOUND_AUDIENCE);
     stop_sample(SOUND_RAIN);
 }
 
-bool step( GameState* s) {
-    Uint8 *keys;
-    SDL_Event e;
+void step(GameState* s) {
     bool ground_event = false;
     int p;
 
+    s->ball.z += s->ball.move_z;
+    if (!s->ball.inhibit_gravity) {
+        s->ball.move_z += GRAVITY;
+    }
+
+    s->ball.x += s->ball.move_x;
+    s->ball.y += s->ball.move_y;
+
+    for (p=1; p<=MAXPLAYERS; p++) {
+        if (PLAYER(s, p).use_power) {
+            PLAYER(s, p).power = fmaxf(0.0, fminf(
+                        PLAYER(s, p).power*POWER_DOWN_FACTOR,
+                        PLAYER_POWER_MAX));
+        }
+    }
+
     if (s->ball.z < 0) {
         /* ground event */
         ground_event = true;
@@ -288,10 +275,8 @@ bool step( GameState* s) {
 
         /* bounce from the ground */
         if (fabsf(s->ball.move_z) > 0.3) {
-            s->sound_events |= SOUND_EVENT_GROUND;
-            sample_volume_group(SOUND_GROUND_FIRST, SOUND_GROUND_LAST, fmaxf(0.0, fminf(1.0, fabsf(s->ball.move_z)/2)));
-            pan_sample_group(SOUND_GROUND_FIRST, SOUND_GROUND_LAST, fmaxf(0.0, fminf(1.0, (s->ball.x)/WIDTH)));
-            s->ball.move_z *= -s->ball.restitution;
+            s->sound_events ^= SOUND_EVENT_GROUND;
+            s->ball.move_z *= -BALL_RESTITUTION;
         } else {
             s->ball.move_z = 0;
         }
@@ -317,25 +302,25 @@ bool step( GameState* s) {
         if (NET_COLLISION_BALL(s->ball)) {
             /* the ball "fell" into the net  */
             s->score_event = SCORE_EVENT_NET;
-            s->sound_events |= SOUND_EVENT_OUT;
-            s->status = "net!";
+            s->sound_events ^= SOUND_EVENT_OUT;
+            s->status_message = STATUSMSG_NET;
         } else if (IS_OFFSCREEN(s->ball.x, s->ball.y)) {
             /* ball flew offscreen */
             s->score_event = SCORE_EVENT_OFFSCREEN;
-            s->sound_events |= SOUND_EVENT_OUT;
-            s->status = "out!";
+            s->sound_events ^= SOUND_EVENT_OUT;
+            s->status_message = STATUSMSG_OUT;
         } else if (ground_event) {
             /* the ball hit the ground on the screen */
             if (IS_OUT(s->ball.x, s->ball.y)) {
                 /* the ball bounced in the OUT area */
                 s->score_event = SCORE_EVENT_OUT;
-                s->sound_events |= SOUND_EVENT_OUT;
-                s->status = "out!";
+                s->sound_events ^= SOUND_EVENT_OUT;
+                s->status_message = STATUSMSG_OUT;
             } else if (GROUND_IS_VALID(s->ball.last_hit_by, s->ball.x, s->ball.y)) {
                 if (s->ball.ground_hit) {
                     s->score_event = SCORE_EVENT_GROUND_VALID;
-                    s->sound_events |= SOUND_EVENT_OUT;
-                    s->status = "did not catch the ball!";
+                    s->sound_events ^= SOUND_EVENT_OUT;
+                    s->status_message = STATUSMSG_DIDNTCATCH;
                 } else {
                     /* first ground hit in valid area */
                     s->ball.ground_hit = true;
@@ -343,27 +328,27 @@ bool step( GameState* s) {
             } else {
                 /* ball hit somewhere invalid */
                 s->score_event = SCORE_EVENT_GROUND_INVALID;
-                s->sound_events |= SOUND_EVENT_OUT;
-                s->status = "fault!";
+                s->sound_events ^= SOUND_EVENT_OUT;
+                s->status_message = STATUSMSG_FAULT;
             }
         }
     }
 
     if (s->score_event != SCORE_UNDECIDED) {
         /* we have some scoring to do */
-        if (s->score_time == 0) {
+        if (s->score_time < SCORING_DELAY/GAME_TICKS) {
             /* schedule scoring in the future */
-            s->score_time = s->time + SCORING_DELAY;
+            s->score_time++;
             s->referee = REFEREE_OUT;
-        } else if (s->time >= s->score_time) {
+        } else if (s->score_time >= SCORING_DELAY/GAME_TICKS) {
             /* time has ran out - score now */
             switch (score_game(s)) {
                 case WINNER_PLAYER1:
-                    s->status = "player 1 scores";
+                    s->status_message = STATUSMSG_P1SCORES;
                     s->referee = REFEREE_PLAYER1;
                     break;
                 case WINNER_PLAYER2:
-                    s->status = "player 2 scores";
+                    s->status_message = STATUSMSG_P2SCORES;
                     s->referee = REFEREE_PLAYER2;
                     break;
                 default:
@@ -372,18 +357,10 @@ bool step( GameState* s) {
             }
             s->score_time = 0;
 
-            strcpy(s->game_score_str, format_game(s));
-            strcpy(s->sets_score_str, format_sets(s));
-            s->text_changed = true;
-
             game_setup_serve(s);
             if (s->location->max_visitors > 100) {
-                s->sound_events |= SOUND_EVENT_APPLAUSE;
+                s->sound_events ^= SOUND_EVENT_APPLAUSE;
             }
-            /*FIXME n-gram predictor broken
-            s->history_size = 0;
-            s->history_is_locked = 0;
-            s->ngram_prediction = 0.0;*/
         }
     } else {
         /* score is still undecided - do the racket swing thing */
@@ -391,9 +368,9 @@ bool step( GameState* s) {
             if (IS_NEAR_X(PLAYER(s, p).x, s->ball.x) && IS_NEAR_Y(PLAYER(s, p).y, s->ball.y-s->ball.z) && PLAYER(s, p).use_power && PLAYER(s, p).power > 30.0 && s->ball.last_hit_by != p) {
                 /* RACKET HIT */
                 if (!s->ball.ground_hit && s->ball.move_x != 0.0) {
-                    s->status = "volley!";
+                    s->status_message = STATUSMSG_VOLLEY;
                 } else {
-                    s->status = format_status(s);
+                    s->status_message = STATUSMSG_DEFAULT;
                 }
                 switch (PLAYER(s, p).desire) {
                     case DESIRE_NORMAL:
@@ -413,168 +390,136 @@ bool step( GameState* s) {
                         break;
                 }
                 s->ball.move_y = get_move_y( s, p);
-                s->sound_events |= SOUND_EVENT_RACKET;
+                s->sound_events ^= SOUND_EVENT_RACKET;
                 s->ball.ground_hit = false;
                 s->ball.inhibit_gravity = false;
                 s->ball.last_hit_by = p;
-                if (p==1) {
-                    pan_sample_group(SOUND_RACKET_FIRST, SOUND_RACKET_LAST, 0.3);
-                } else {
-                    pan_sample_group(SOUND_RACKET_FIRST, SOUND_RACKET_LAST, 0.7);
+                if (p==2) {
                     s->ball.move_x *= -1;
                 }
             }
         }
     }
+}
 
-#if 0
-    FIXME n-gram predictor broken
-    if( IS_OUT_X(s->ball.x) || IS_OFFSCREEN_Y(s->ball.y)) {
-        if( IS_OUT_X(s->ball.x)) {
-            /*if( !s->history_is_locked && s->referee != REFEREE_OUT) {
-                s->history[s->history_size] = (int)(NGRAM_STEPS*s->ball.y/HEIGHT);
-                s->history_size++;
-                if( s->history_size == 3) {
-                    s->ngram[s->history[0]][s->history[1]][s->history[2]] += 10;
-                    s->ngram_prediction = ngram_predictor( s);
-                    s->history[0] = s->history[1];
-                    s->history[1] = s->history[2];
-                    s->history_size--;
-                }
-                s->history_is_locked = true;
-            }*/
-        }
-    } else {
-        /*s->history_is_locked = false;*/
-    }
-#endif
+bool handle_input(GameState* s, TennixNet* c) {
+    static NetworkGameState tmp;
+    static NetworkInputData net_input;
+    Uint8* keys = NULL;
+    int p;
 
     SDL_PumpEvents();
     keys = SDL_GetKeyState(NULL);
 
-    if( s->time%50==0) {
-        if (keys['r']) {
-            if (s->rain == 0) {
-                play_sample_background(SOUND_RAIN);
-            }
-            s->rain += 10;
-        }
-        if (keys['t']) {
-            s->fog++;
-        }
-        if (keys['1']) {
-            s->wind++;
-        }
-        if (keys['2']) {
-            s->wind--;
-        }
+    if (keys['1']) {
+        net_serialize_gamestate(s, &tmp);
+    } else if (keys['2']) {
+        net_unserialize_gamestate(&tmp, s);
     }
 
-    if(!(SDL_GetTicks() < fading_start+FADE_DURATION) && !s->is_over) {
+    network_get_input(c, &net_input);
+    if (s->winner == WINNER_NONE) {
         for (p=1; p<=MAXPLAYERS; p++) {
             if( PLAYER(s, p).type == PLAYER_TYPE_HUMAN) {
+                if (PLAYER(s, p).input->type == INPUT_TYPE_NETWORK) {
+                    memcpy(&(PLAYER(s, p).input->net), &net_input,
+                            sizeof(NetworkInputData));
+                }
                 input_human(s, p);
+                if (PLAYER(s, p).input->type != INPUT_TYPE_NETWORK) {
+                    network_send_input(c, &(PLAYER(s, p).input->net));
+                }
             } else {
                 input_ai(s, p);
             }
+
+            /* Make sure player coordinates are valid */
+            if (PLAYER(s, p).y < PLAYER_Y_MIN) {
+                PLAYER(s, p).y = PLAYER_Y_MIN;
+            } else if (PLAYER(s, p).y > PLAYER_Y_MAX) {
+                PLAYER(s, p).y = PLAYER_Y_MAX;
+            }
         }
     }
     
-    /* Maemo: The "F6" button is the "Fullscreen" button */
-    /*if( keys['f'] || keys[SDLK_F6]) SDL_WM_ToggleFullScreen( screen);*/
-    if( keys['y']) SDL_SaveBMP( screen, "screenshot.bmp");
-
     /* Maemo: The "F4" button is the "Open menu" button */
-    if( keys['p'] || keys[SDLK_F4]) {
-        while( keys['p'] || keys[SDLK_F4]) {
-            SDL_PollEvent( &e);
-            keys = SDL_GetKeyState( NULL);
-            SDL_Delay( 10);
-        }
-        while( (keys['p'] || keys[SDLK_F4]) == 0) {
-            SDL_PollEvent( &e);
-            keys = SDL_GetKeyState( NULL);
-            SDL_Delay( 10);
-        }
-        while( keys['p'] || keys[SDLK_F4]) {
-            SDL_PollEvent( &e);
-            keys = SDL_GetKeyState( NULL);
-            SDL_Delay( 10);
-        }
-        s->was_stopped = true;
-    }
-     
-    if( keys[SDLK_ESCAPE] || keys['q']) return true;
-    
-    limit_value( &PLAYER(s, 1).y, PLAYER_Y_MIN, PLAYER_Y_MAX);
-    limit_value( &PLAYER(s, 2).y, PLAYER_Y_MIN, PLAYER_Y_MAX);
-    
-    if (s->ball.move_x > 0 && s->wind > 0) {
-        s->ball.x += fabsf(s->wind)/2;
-    } else if (s->ball.move_x < 0 && s->wind < 0) {
-        s->ball.x -= fabsf(s->wind)/2;
-    }
-
-    s->ball.z += s->ball.move_z;
-    if (!s->ball.inhibit_gravity) {
-        s->ball.move_z += GRAVITY;
-    }
-
-    s->ball.x += s->ball.move_x;
-    s->ball.y += s->ball.move_y;
-
-    for (p=1; p<=MAXPLAYERS; p++) {
-        if (PLAYER(s, p).use_power) {
-            PLAYER(s, p).power = fmaxf(0.0, fminf(PLAYER(s, p).power*PLAYER(s, p).power_down_factor, PLAYER_POWER_MAX));
-        }
-    }
-
-    return false;
+    return (keys[SDLK_ESCAPE] || keys['q']);
 }
 
-void render( GameState* s) {
+void render(const GameState* s, RenderState* r) {
     int x, y, b;
     unsigned int i;
     float zoom;
     float rotate;
     int t=1000;
+    soundevent_t sounds;
+    Uint32 time = SDL_GetTicks();
 
-    if (s->sound_events != 0) {
-        if (s->sound_events & SOUND_EVENT_GROUND) {
+    /* The bits in sound_events flip when the sound should play */
+    if ((sounds = (r->sound_events ^ s->sound_events)) != 0) {
+        if (sounds & SOUND_EVENT_GROUND) {
+            sample_volume_group(SOUND_GROUND_FIRST, SOUND_GROUND_LAST,
+                    fmaxf(0.0, fminf(1.0, fabsf(s->ball.move_z)/2)));
+            pan_sample_group(SOUND_GROUND_FIRST, SOUND_GROUND_LAST,
+                    fmaxf(0.0, fminf(1.0, (s->ball.x)/WIDTH)));
             play_sample(SOUND_GROUND);
         }
-        if (s->sound_events & SOUND_EVENT_OUT) {
+        if (sounds & SOUND_EVENT_OUT) {
             play_sample(SOUND_OUT);
         }
-        if (s->sound_events & SOUND_EVENT_APPLAUSE) {
+        if (sounds & SOUND_EVENT_APPLAUSE) {
             play_sample(SOUND_APPLAUSE);
         }
-        if (s->sound_events & SOUND_EVENT_RACKET) {
+        if (sounds & SOUND_EVENT_RACKET) {
+            if (((s->ball.x)/WIDTH) < 0.5) {
+                pan_sample_group(SOUND_RACKET_FIRST, SOUND_RACKET_LAST, 0.3);
+            } else {
+                pan_sample_group(SOUND_RACKET_FIRST, SOUND_RACKET_LAST, 0.7);
+            }
             play_sample(SOUND_RACKET);
         }
-        s->sound_events = 0;
+        r->sound_events = s->sound_events;
     }
+
     if( s->winner != WINNER_NONE) {
-        if( !s->is_over) {
-            start_fade();
-            s->is_over = true;
-        }
         clear_screen();
-        rectangle(0, 0, WIDTH, HEIGHT, 80, 80, 80);
+        //rectangle(0, 0, WIDTH, HEIGHT, 80, 80, 80);
         store_screen();
         show_sprite( GR_RACKET, 2*(s->winner-1), 4, WIDTH/2 - get_image_width( GR_RACKET)/8, HEIGHT/2 - get_image_height( GR_RACKET), 255);
-        sprintf( s->game_score_str, "player %d wins the match with %s", s->winner, format_sets( s));
-        font_draw_string(FONT_LARGE, s->game_score_str, (WIDTH-font_get_string_width(FONT_LARGE, s->game_score_str))/2, HEIGHT/2 + 30);
+        /*sprintf( s->game_score_str, "player %d wins the match with %s", s->winner, format_sets( s));
+        font_draw_string(FONT_LARGE, s->game_score_str, (WIDTH-font_get_string_width(FONT_LARGE, s->game_score_str))/2, HEIGHT/2 + 30);*/
         updatescr();
         return;
     }
-    if (s->displayed_court_type != s->location->court_type || s->text_changed || s->old_referee != s->referee) {
+    if ((r->referee != s->referee) ||
+            (r->ec_game != s->ec_game) ||
+            (r->ec_sets != s->ec_sets) ||
+            (r->status_message != s->status_message)) {
+        /* Update status message text */
+        if (r->status_message != s->status_message) {
+            r->text_status = format_status(s);
+            r->status_message = s->status_message;
+        }
+        /* Update game status text */
+        if (r->ec_game != s->ec_game) {
+            r->text_game = format_game(s);
+            r->ec_game = s->ec_game;
+        }
+        /* Update set status text */
+        if (r->ec_sets != s->ec_sets) {
+            r->text_sets = format_sets(s);
+            if (r->status_message == STATUSMSG_DEFAULT) {
+                r->text_status = format_status(s);
+            }
+            r->ec_sets = s->ec_sets;
+        }
         clear_screen();
-        rectangle(0, 0, WIDTH, HEIGHT, 80, 80, 80);
+        //rectangle(0, 0, WIDTH, HEIGHT, 80, 80, 80);
         fill_image(s->location->court_type, 120, 120, 400, 250);
         show_image(GR_COURT, 0, 0, 255);
-        font_draw_string(FONT_XLARGE, s->game_score_str, 14, 14);
-        font_draw_string(FONT_XLARGE, s->sets_score_str, (WIDTH-font_get_string_width(FONT_XLARGE, s->sets_score_str))-14, 14);
+        font_draw_string(FONT_XLARGE, r->text_game, 14, 14);
+        font_draw_string(FONT_XLARGE, r->text_sets,
+                (WIDTH-font_get_string_width(FONT_XLARGE, r->text_sets))-14, 14);
         if (s->location->has_referee) {
             switch (s->referee) {
                 case REFEREE_NORMAL:
@@ -588,7 +533,7 @@ void render( GameState* s) {
                     t = 400;
                     break;
             }
-            t = (s->time/t)%4;
+            t = (time/t)%4;
             switch (t) {
                 case 0:
                     t=0;
@@ -605,12 +550,10 @@ void render( GameState* s) {
             }
             show_sprite( GR_REFEREE, s->referee*3+t, 12, 250, 10, 255);
             if (voice_finished_flag == 0) {
-                show_sprite(GR_TALK, (s->time/150)%2, 2, 280, 45, 255);
+                show_sprite(GR_TALK, (time/150)%2, 2, 280, 45, 255);
             }
         }
-        s->displayed_court_type = s->location->court_type;
-        s->text_changed = false;
-        s->old_referee = s->referee;
+        r->referee = s->referee;
         store_screen();
     }
     
@@ -626,9 +569,9 @@ void render( GameState* s) {
     rectangle(WIDTH-PLAYER_POWER_MAX-10, HEIGHT-30, (int)(PLAYER(s, 2).power), 10, 200, 200, 200);
     
     if( s->ball.move_x > 0) {
-        b = (s->time/100)%BALL_STATES;
+        b = (time/100)%BALL_STATES;
     } else if( s->ball.move_x < 0) {
-        b = BALL_STATES-1-(s->time/100)%BALL_STATES;
+        b = BALL_STATES-1-(time/100)%BALL_STATES;
     } else {
         b = 0;
     }
@@ -637,7 +580,7 @@ void render( GameState* s) {
     zoom = fmaxf(0.5, fminf(1.0, (float)(30.-(s->ball.z))/10.));
     show_image_rotozoom(GR_SHADOW, s->ball.x, s->ball.y+3, rotate, zoom);
 
-    rotate = (-s->ball.move_x * (float)((float)s->time/10.));
+    rotate = (-s->ball.move_x * (float)((float)time/10.));
     zoom = 1.0 + fmaxf(0.0, (s->ball.z-10.)/100.);
     show_image_rotozoom(GR_BALL, s->ball.x, s->ball.y - s->ball.z, rotate, zoom);
 
@@ -653,14 +596,15 @@ void render( GameState* s) {
         rectangle(PLAYER(s, 2).x-1+5, PLAYER(s, 2).input->my-1, 2, 2, 0, 0, 0);
     }
 
-    font_draw_string(FONT_MEDIUM, s->status, (WIDTH-font_get_string_width(FONT_MEDIUM, s->status))/2, HEIGHT-50);
+    font_draw_string(FONT_MEDIUM, r->text_status,
+            (WIDTH-font_get_string_width(FONT_MEDIUM, r->text_status))/2, HEIGHT-50);
 
-    for (i=0; i<s->rain; i++) {
+    for (i=0; i<s->location->rain; i++) {
         x = rand()%WIDTH;
         y = rand()%HEIGHT;
-        draw_line_faded(x, y, x+10+s->wind*5, y+30, 0, 0, 255, 100, 200, 255);
+        draw_line_faded(x, y, x+10, y+30, 0, 0, 255, 100, 200, 255);
     }
-    if (s->rain) {
+    if (s->location->rain) {
         /**
          * Cheap-ish update of the whole screen. This can
          * probably be optimized.
@@ -680,34 +624,26 @@ void render( GameState* s) {
     line_horiz( GAME_Y_MIN, 100, 100, 100);
     line_horiz( GAME_Y_MAX, 100, 100, 100);
 #endif
-    switch (s->fog) {
+    switch (s->location->fog) {
         default:
         case 4:
-            fill_image_offset(GR_FOG2, 0, 0, WIDTH, HEIGHT, -s->time/150-s->windtime/200, 0);
+            fill_image_offset(GR_FOG2, 0, 0, WIDTH, HEIGHT, -time/150, 0);
         case 3:
-            fill_image_offset(GR_FOG, 0, 0, WIDTH, HEIGHT, -s->time/100-s->windtime/150, 20);
+            fill_image_offset(GR_FOG, 0, 0, WIDTH, HEIGHT, -time/100, 20);
         case 2:
-            fill_image_offset(GR_FOG2, 0, 0, WIDTH, HEIGHT, -s->time/180-s->windtime/180, 80);
+            fill_image_offset(GR_FOG2, 0, 0, WIDTH, HEIGHT, -time/180, 80);
         case 1:
-            fill_image_offset(GR_FOG, 0, 0, WIDTH, HEIGHT, s->time/200-s->windtime/100, 0);
+            fill_image_offset(GR_FOG, 0, 0, WIDTH, HEIGHT, time/200, 0);
         case 0:
             break;
     }
-    if (s->night) {
+    if (s->location->night) {
         show_image(GR_NIGHT, 0, 0, 255);
     }
 
     updatescr();
 }
 
-void limit_value( float* value, float min, float max) {
-    if( *value < min) {
-        *value = min;
-    } else if( *value > max) {
-        *value = max;
-    }
-}
-
 float get_move_y( GameState* s, unsigned char player) {
     float pct, dest, x_len, y_len;
     float py, by, pa, move_x;
@@ -729,7 +665,7 @@ float get_move_y( GameState* s, unsigned char player) {
     } else {
         x_len = s->ball.x - GAME_X_MIN;
     }
-    y_len = dest - by + MOVE_Y_SEED-rand()%MOVE_Y_SEED*2;
+    y_len = dest - by;
 
     /* return the should-be value for move_y */
     return (y_len*move_x)/(x_len);
@@ -757,6 +693,9 @@ void input_human(GameState* s, int player) {
         }
         PLAYER(s, player).y += move_y;
         PLAYER(s, player).accelerate *= PLAYER_ACCEL_INCREASE;
+        if (PLAYER(s, player).accelerate > 190) {
+            PLAYER(s, player).accelerate = 190;
+        }
     } else {
         PLAYER(s, player).accelerate = PLAYER_ACCEL_DEFAULT;
     }
@@ -765,7 +704,7 @@ void input_human(GameState* s, int player) {
         PLAYER(s, player).desire = (topspin)?(DESIRE_TOPSPIN):(DESIRE_NORMAL);
         PLAYER(s, player).desire = (smash)?(DESIRE_SMASH):(PLAYER(s, player).desire);
 
-        PLAYER(s, player).power = fmaxf(10.0, fminf(PLAYER(s, player).power*PLAYER(s, player).power_up_factor, PLAYER_POWER_MAX));
+        PLAYER(s, player).power = fmaxf(10.0, fminf(PLAYER(s, player).power*POWER_UP_FACTOR, PLAYER_POWER_MAX));
         PLAYER(s, player).use_power = false;
     } else {
         PLAYER(s, player).use_power = true;
@@ -774,7 +713,6 @@ void input_human(GameState* s, int player) {
 
 void input_ai(GameState* s, int player) {
     float fact = 1.7;
-    float target;
     int ball_approaching = 0;
 
     if ((PLAYER(s, player).x < GAME_X_MID && s->ball.move_x <= 0) ||
@@ -800,20 +738,9 @@ void input_ai(GameState* s, int player) {
         if (IS_NEAR_Y(PLAYER(s, player).y, (s->ball.y-s->ball.z)) && IS_NEAR_X_AI(PLAYER(s, player).x, s->ball.x) && PLAYER(s, player).power > 90.) {
             PLAYER(s, player).use_power = true;
         } else if (ball_approaching) {
-            PLAYER(s, player).power = fmaxf(10., fminf(PLAYER(s, player).power*PLAYER(s, player).power_up_factor, PLAYER_POWER_MAX));
+            PLAYER(s, player).power = fmaxf(10., fminf(PLAYER(s, player).power*POWER_UP_FACTOR, PLAYER_POWER_MAX));
             PLAYER(s, player).use_power = false;
         }
-    } else if( s->ngram_prediction > 0.0) {
-        target = s->ngram_prediction*((float)HEIGHT)/((float)(NGRAM_STEPS));
-        target = GAME_Y_MID + (target-GAME_Y_MID)*1.5;
-
-        if( PLAYER(s, player).desire == DESIRE_NORMAL && !IS_NEAR_Y_AI( PLAYER(s, player).y, target)) {
-            if( PLAYER(s, player).y < target) {
-                PLAYER(s, player).y += fminf(fact, (target-PLAYER(s, player).y)/40.0);
-            } else if( PLAYER(s, player).y > target) {
-                PLAYER(s, player).y -= fminf(fact, (PLAYER(s, player).y-target)/40.0);
-            }
-        }
     }
 }
 
@@ -833,32 +760,6 @@ void game_setup_serve( GameState* s) {
     }
 }
 
-float ngram_predictor( GameState* s) {
-    unsigned int count = 0;
-    unsigned long sum = 0;
-    int x, y, z;
-    float result;
-
-    if( s->history_size < 3) {
-        return 0.0;
-    }
-
-    x = s->history[1];
-    y = s->history[2];
-
-    for( z = 0; z<NGRAM_STEPS; z++) {
-        count += s->ngram[x][y][z];
-        sum += z * s->ngram[x][y][z];
-    }
-
-    result = ((float)(sum))/((float)(count));
-#ifdef DEBUG
-    printf( "predicting next = %.2f\n", result);
-#endif
-
-    return result;
-}
-
 int score_game(GameState* s) {
     Player *last_hit_by = NULL;
     Player *other = NULL;
@@ -904,10 +805,12 @@ int score_game(GameState* s) {
     assert(s->current_set < SETS_TO_WIN*2-1);
 
     winner->game++;
+    s->ec_game++;
     if( loser->game < winner->game-1) {
         if( winner->game >= 4) {
             winner->game = loser->game = 0;
             winner->sets[s->current_set]++;
+            s->ec_sets++;
 
             /* serving is changed when the "game" is over */
             s->serving_player = (s->serving_player==1)?(2):(1);
@@ -936,7 +839,7 @@ int score_game(GameState* s) {
     }
 }
 
-char* format_sets( GameState* s) {
+const char* format_sets(const GameState* s) {
     static char sets[100];
     static char tmp[100];
     int i, max = s->current_set;
@@ -956,7 +859,7 @@ char* format_sets( GameState* s) {
     return sets;
 }
 
-char* format_game( GameState* s) {
+const char* format_game(const GameState* s) {
     static char game[100];
     static const int game_scoring[] = { 0, 15, 30, 40 };
 
@@ -991,16 +894,41 @@ char* format_game( GameState* s) {
     return game;
 }
 
-char* format_status( GameState* s) {
+const char* format_status(const GameState* s) {
     static char status[100];
     static const char* set_names[] = { "first", "second", "third", "fourth", "fifth" };
 
-    sprintf( status, "%d:%d in %s set", PLAYER(s, 1).sets[s->current_set], PLAYER(s, 2).sets[s->current_set], set_names[s->current_set]);
-
-    return status;
+    switch (s->status_message) {
+        case STATUSMSG_NONE:
+            return "";
+        case STATUSMSG_WELCOME:
+            return "welcome to tennix " VERSION;
+        case STATUSMSG_DEFAULT:
+            sprintf(status, "%d:%d in %s set",
+                    PLAYER(s, 1).sets[s->current_set],
+                    PLAYER(s, 2).sets[s->current_set],
+                    set_names[s->current_set]);
+            return status;
+        case STATUSMSG_NET:
+            return "net!";
+        case STATUSMSG_OUT:
+            return "out!";
+        case STATUSMSG_FAULT:
+            return "fault!";
+        case STATUSMSG_P1SCORES:
+            return "player 1 scores.";
+        case STATUSMSG_P2SCORES:
+            return "player 2 scores.";
+        case STATUSMSG_VOLLEY:
+            return "volley!";
+        case STATUSMSG_DIDNTCATCH:
+            return "did not catch the ball.";
+        default:
+            return "";
+    }
 }
 
-int game_get_winner( GameState* s) {
+winner_t game_get_winner(const GameState* s) {
     unsigned int i;
     int sets[2] = {0};
 
