@@ -30,9 +30,8 @@
 #include "graphics.h"
 #include "input.h"
 
-#define SETS_TO_WIN 3
 
-#define NGRAM_STEPS 6
+#define SETS_TO_WIN 3
 
 #define MAXPLAYERS 2
 
@@ -42,12 +41,7 @@
 #  define BALL_STATES 4
 #endif
 
-typedef unsigned char bool;
-enum {
-    false,
-    true
-};
-
+typedef unsigned char referee_t;
 enum {
     REFEREE_NORMAL,
     REFEREE_OUT,
@@ -56,18 +50,22 @@ enum {
     REFEREE_COUNT
 };
 
+typedef unsigned char winner_t;
 enum {
     WINNER_NONE,
     WINNER_PLAYER1,
     WINNER_PLAYER2
 };
 
+typedef unsigned char soundevent_t;
 enum {
+    SOUND_EVENT_NONE     = 0<<0,
     SOUND_EVENT_GROUND   = 1<<0,
     SOUND_EVENT_OUT      = 1<<1,
     SOUND_EVENT_APPLAUSE = 1<<2,
     SOUND_EVENT_RACKET   = 1<<3
 };
+#define SOUND_EVENT_COUNT 4
 
 typedef struct {
     const char* name;
@@ -81,8 +79,13 @@ typedef struct {
     unsigned int worldmap_x;
     unsigned int worldmap_y;
     bool has_referee;
+    bool night;
+    unsigned int rain;
+    unsigned int fog;
 } Location;
 
+#define BALL_RESTITUTION .6
+
 typedef struct {
     float x;
     float y;
@@ -90,32 +93,24 @@ typedef struct {
     float move_x;
     float move_y;
     float move_z;
-    float restitution;
     bool ground_hit;
-    int last_hit_by;
+    char last_hit_by;
     bool inhibit_gravity;
 } Ball;
 
 typedef struct {
     InputDevice* input;
-    int input_device_index;
+    char input_device_index;
     float x;
     float y;
     float power;
-    float power_up_factor;
-    float power_down_factor;
     bool use_power;
-    unsigned int score;
+    unsigned char score;
     unsigned char desire;
     bool type; /* is this player ai-controlled or human? */
-    float ball_dest; /* prospective y-position of ball */
-    bool state_locked; /* enabled when user keeps pressing the "hit" key */
-    int game; /* score for the current game */
-    int sets[SETS_TO_WIN*2]; /* score for each set */
-    int mouse_x; /* x position of mouse */
-    int mouse_y; /* y position of mouse */
+    char game; /* score for the current game */
+    unsigned char sets[SETS_TO_WIN*2]; /* score for each set */
     float accelerate; /* a value [0..1] how fast the user accelerates */
-    bool mouse_locked; /* on start, ignore unpressed mouse state */
 } Player;
 
 enum {
@@ -131,8 +126,9 @@ enum {
 };
 
 /* wait 2 seconds before we score the game */
-#define SCORING_DELAY 2000
+#define SCORING_DELAY 1000
 
+typedef unsigned char scoreevent_t;
 enum {
     SCORE_UNDECIDED,
     SCORE_EVENT_NET,
@@ -143,43 +139,54 @@ enum {
     SCORE_EVENT_MAX
 };
 
+typedef unsigned char eventcounter_t;
+enum {
+    EVENTCOUNTER_RENDERSTATE_START = 0,
+    EVENTCOUNTER_GAMESTATE_START = 1
+};
+
+typedef unsigned char statusmessage_t;
+enum {
+    STATUSMSG_NONE,
+    STATUSMSG_WELCOME,
+    STATUSMSG_DEFAULT,
+    STATUSMSG_NET,
+    STATUSMSG_OUT,
+    STATUSMSG_FAULT,
+    STATUSMSG_P1SCORES,
+    STATUSMSG_P2SCORES,
+    STATUSMSG_VOLLEY,
+    STATUSMSG_DIDNTCATCH
+};
+
 typedef struct {
-    Location* location;
-    int current_location; /* index of loc. in global location table */
+    const Location* location;
+    char current_location; /* index of loc. in global location table */
     Ball ball;
     Player players[MAXPLAYERS];
-    Uint32 time;
-    bool was_stopped;
-    unsigned int serving_player;
-    const char* status;
-    char game_score_str[50];
-    char sets_score_str[50];
-    unsigned char referee;
-    unsigned int current_set;
-    int winner;
-    bool is_over;
-    image_id displayed_court_type;
-    unsigned int history[3];
-    unsigned int history_size;
-    bool history_is_locked;
-    unsigned char ngram[NGRAM_STEPS][NGRAM_STEPS][NGRAM_STEPS];
-    float ngram_prediction;
-    unsigned int sound_events;
-    float joystick_y;
-    float joystick_x;
-    unsigned char joystick_a;
-    unsigned int rain;
-    unsigned int fog;
-    bool night;
-    int wind;
-    unsigned int windtime;
-    Uint32 timelimit;
-    bool text_changed;
-    unsigned char old_referee;
-    int score_event;
-    unsigned int score_time;
+    unsigned char serving_player;
+    referee_t referee;
+    unsigned char current_set;
+    winner_t winner;
+    soundevent_t sound_events;
+    scoreevent_t score_event;
+    unsigned char score_time;
+    eventcounter_t ec_game;
+    eventcounter_t ec_sets;
+    statusmessage_t status_message;
 } GameState;
 
+typedef struct {
+    soundevent_t sound_events;
+    referee_t referee;
+    eventcounter_t ec_game;
+    eventcounter_t ec_sets;
+    statusmessage_t status_message;
+    const char* text_status;
+    const char* text_game;
+    const char* text_sets;
+} RenderState;
+
 #define PI 3.1415
 
 #define GAME_TICKS 15
@@ -255,8 +262,6 @@ typedef struct {
 #define POWER_DOWN_FACTOR 0.9
 #define PLAYER_POWER_MAX 100
 
-#define MOVE_Y_SEED 3
-
 #define PLAYER(s, num) (s->players[num-1])
 
 /* Comment out the following #define to enable mouse control */
@@ -269,21 +274,22 @@ int gamestate_save(GameState* s, const char* filename);
 GameState* gamestate_load(const char* filename);
 
 /* Game module functions */
-void gameloop(GameState*);
-void render( GameState*);
-bool step( GameState*);
-void limit_value( float*, float, float);
+void gameloop(GameState*, TennixNet*);
+void step(GameState*);
+bool handle_input(GameState*, TennixNet*);
+void render(const GameState*, RenderState*);
 float get_move_y( GameState*, unsigned char);
 void input_human(GameState*, int);
 void input_ai(GameState*, int);
 void game_setup_serve( GameState*);
-float ngram_predictor( GameState*);
 
 int score_game(GameState*);
-char* format_sets( GameState*);
-char* format_game( GameState*);
-char* format_status( GameState*);
-int game_get_winner( GameState*);
+winner_t game_get_winner(const GameState*);
+
+/* Helper functions for the renderer */
+const char* format_sets(const GameState*);
+const char* format_game(const GameState*);
+const char* format_status(const GameState*);
 
 #endif
 
