@@ -24,6 +24,7 @@
 #ifndef __GRAPHICS_H
 #define __GRAPHICS_H
 
+#include "archive.hh"
 #include "tennix.h"
 
 #include "SDL_image.h"
@@ -42,7 +43,7 @@
 #define SET_PIXEL_RGB(surface,x,y,r,g,b) (GET_PIXEL_DATA(surface,x,y)=SDL_MapRGB(surface->format, r, g, b))
 
 struct _FreeListItem {
-    char* data;
+    const char* data;
     struct _FreeListItem* next;
 };
 
@@ -53,7 +54,7 @@ typedef struct {
 } FreeList;
 
 FreeList* freelist_create();
-void freelist_append(FreeList* list, char* data);
+void freelist_append(FreeList* list, const char* data);
 void freelist_free_all(FreeList* list);
 
 typedef struct {
@@ -126,7 +127,7 @@ enum {
 #define GR_CTT_FIRST GR_CTT_HARD
 #define GR_CTT_LAST GR_CTT_GRASS
 
-void init_graphics();
+void init_graphics(TennixArchive& tnxar);
 void uninit_graphics();
 int get_image_width(image_id);
 int get_image_height(image_id);
